"""
The `~certbot_dns_inwx.dns_inwx` plugin automates the process of
completing a ``dns-01`` challenge (`~acme.challenges.DNS01`) by creating, and
subsequently removing, TXT records using the INWX XML-RPC DNS API.

For further documentation see the README file.
"""
